// SPDX-FileCopyrightText: GSConnect Developers https://github.com/GSConnect
//
// SPDX-License-Identifier: GPL-2.0-or-later

import * as battery from './battery.js';
import * as clipboard from './clipboard.js';
import * as connectivity_report from './connectivity_report.js';
import * as contacts from './contacts.js';
import * as findmyphone from './findmyphone.js';
import * as mousepad from './mousepad.js';
import * as mpris from './mpris.js';
import * as notification from './notification.js';
import * as ping from './ping.js';
import * as presenter from './presenter.js';
import * as runcommand from './runcommand.js';
import * as sftp from './sftp.js';
import * as share from './share.js';
import * as sms from './sms.js';
import * as systemvolume from './systemvolume.js';
import * as telephony from './telephony.js';

export default {
    battery,
    clipboard,
    connectivity_report,
    contacts,
    findmyphone,
    mousepad,
    mpris,
    notification,
    ping,
    presenter,
    runcommand,
    sftp,
    share,
    sms,
    systemvolume,
    telephony,
};
