/*
 * SPDX-FileCopyrightText: GSConnect Developers https://github.com/GSConnect
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/**
 * Chrome/Firefox buttons in Settings Window
 */
.badge-button {
  border: 0px;
  padding: 0px;
  background: transparent;
}


/**
 * Placeholders
 */
.placeholder {
  background: @content_view_bg_color;
  margin-bottom: 30px;
}

.placeholder-image {
  margin-bottom: 16px;
  opacity: 0.5;
}

.placeholder-title {
  font-size: larger;
  font-weight: bold;
  margin-bottom: 3px;
  opacity: 0.5;
}

.placeholder-description {
  font-size: smaller;
  opacity: 0.5;
}


/**
 * GSConnectContactChooser
 */
.contact-window {
  border-top: 1px solid @borders;
}

.contact-list row {
  padding: 0px;
}


/**
 * GSConnectConversationWidget
 */
.message-scrolled {
  border-bottom: 1px solid @borders;
}

/**
 * GSConnectMessagingWindow
 */
.message-list > label {
  font-size: small;
  margin-bottom: 1em;
}

.thread-list {
  background: @content_view_bg_color;
}

/**
 * Incoming Message Bubbles (GtkLabel subclass)
 */
.message-in {
  color: @theme_fg_color;
  background: alpha(@theme_selected_bg_color, 0.2);
  border: 1px solid alpha(@theme_selected_bg_color, 0.2);
  border-radius: 5px;
  box-shadow: 2px 2px 3px -2px rgba(0, 0, 0, 0.75);
  padding: 6px 9px 6px 9px;
}

.message-in selection {
  color: @theme_selected_bg_color;
  background-color: @theme_selected_fg_color;
}


/**
 * Outgoing Message Bubbles (GtkLabel subclass)
 */
.message-out {
  color: @theme_fg_color;
  background-color: @theme_bg_color;
  border: 1px solid alpha(@theme_fg_color, 0.1);
  border-radius: 5px;
  box-shadow: 2px 2px 3px -2px rgba(0, 0, 0, 0.75);
  caret-color: currentColor;
  -gtk-secondary-caret-color: currentColor;
  padding: 6px 9px 6px 9px;
}

.message-out selection {
  background-color: @theme_selected_bg_color;
  color: @theme_selected_fg_color;
}

/**
 * Must contrast with message background
 */
.message-in *:link,
.message-out *:link {
  color: @theme_fg_color;
}

/**
 * Special case for pending messages
 */
.message-pending .message-out {
  opacity: 0.5;
}

/**
 * Error Dialog
 */
.error-stack-frame > * {
  border-top-width: 0;
}

