// SPDX-FileCopyrightText: GSConnect Developers https://github.com/GSConnect
//
// SPDX-License-Identifier: GPL-2.0-or-later


export default class Component {
    clickPointer() {}

    doubleclickPointer() {}

    pressPointer() {}

    releasePointer() {}

    movePointer() {}

    scrollPointer() {}

    pressKeys() {}
}

